use std::{
    borrow::BorrowMut, collections::HashMap, fmt::Debug, fs::File, mem::size_of, ops::Range,
    sync::Arc,
};

mod allocator;
mod atomic_arc;

use allocator::{AllocatorState, FreeList, GeneralPurposeAllocator, SlabListPointer, SlabPointer};
use atomic_arc::AtomicArc;
use memmap::{Mmap, MmapMut};
use zerocopy::{AsBytes, FromBytes, LayoutVerified, Unaligned, LE};

const PAGE_SIZE: u64 = 4096;

type U64 = zerocopy::byteorder::U64<LE>;
type U32 = zerocopy::byteorder::U32<LE>;
type U16 = zerocopy::byteorder::U16<LE>;

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned, Hash, PartialEq, Eq)]
#[repr(transparent)]
pub struct FilePointer(U64);

impl PartialOrd for FilePointer {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        self.0.get().partial_cmp(&other.0.get())
    }
}

impl Ord for FilePointer {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        self.0.get().cmp(&other.0.get())
    }
}

impl Debug for FilePointer {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "0x{:x}", self.0.get())
    }
}

impl FilePointer {
    fn page(self) -> PagePointer {
        PagePointer(u32::try_from(self.0.get() / PAGE_SIZE).unwrap().into())
    }
    fn page_offset(self) -> (PagePointer, u16) {
        (self.page(), (self.0.get() % PAGE_SIZE) as u16)
    }
    fn from_page_offset(page: PagePointer, offset: u16) -> Self {
        debug_assert!(
            offset < PAGE_SIZE as u16,
            "offset 0x{offset:x} out for page bounds (0..0x{PAGE_SIZE:x})"
        );
        page.start() + offset as u64
    }
    fn null() -> Self {
        Self(U64::ZERO)
    }
    fn is_null(self) -> bool {
        self == Self::null()
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned, Debug, PartialEq, Eq)]
#[repr(transparent)]
pub struct PagePointer(U32);

impl PagePointer {
    fn start(self) -> FilePointer {
        FilePointer((self.0.get() as u64 * PAGE_SIZE).into())
    }
    fn range(self) -> FileRange {
        self.start().range(PAGE_SIZE)
    }
    fn nth(n: u32) -> Self {
        Self(n.into())
    }
    fn null() -> Self {
        Self::nth(0)
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned, PartialEq, Eq)]
#[repr(C)]
pub struct FileRange {
    start: FilePointer,
    len: U64,
}

impl Debug for FileRange {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{:?}..{:?}", self.start, self.start + self.len.get())
    }
}

impl std::ops::Add<u64> for FilePointer {
    type Output = Self;

    fn add(self, rhs: u64) -> Self::Output {
        Self((self.0.get() + rhs).into())
    }
}

impl FilePointer {
    pub fn range(&self, len: u64) -> FileRange {
        FileRange {
            start: *self,
            len: U64::new(len),
        }
    }
}

impl FileRange {
    fn len(&self) -> u64 {
        self.len.get()
    }
    fn as_range(&self) -> Range<usize> {
        let start = self.start.0.get().try_into().unwrap();
        start..start + usize::try_from(self.len()).unwrap()
    }
}

#[derive(Clone, Copy, FromBytes, AsBytes, Unaligned)]
#[repr(C)]
struct Header {
    magic: [u8; 16],
    root: FilePointer,
    allocator_state: AllocatorState,
}

impl Default for Header {
    fn default() -> Self {
        Self {
            magic: *b"cool db format 1",
            root: FilePointer::null(),
            allocator_state: AllocatorState {
                general: FilePointer::null(),
                slabs: SlabListPointer(FilePointer::null() + size_of::<Header>() as u64),
            },
        }
    }
}

struct Snapshot {
    root: FilePointer,
    map: Mmap,
}

pub struct Reader {
    state: Arc<AtomicArc<Snapshot>>,
}

pub struct Db {
    file: File,
    map: MmapMut,
    state: Arc<AtomicArc<Snapshot>>,
}

#[derive(Clone, Copy)]
struct Replaced {
    from: FileRange,
    to: Option<FileRange>,
}

pub struct TransactionHandle<'t> {
    db: &'t mut Db,
    replaced: HashMap<FilePointer, Replaced>,
    new: HashMap<FilePointer, FileRange>,
}

impl<'t> TransactionHandle<'t> {
    unsafe fn get_data(&mut self, range: FileRange) -> &mut [u8] {
        &mut self.db.map[range.as_range()]
    }

    pub unsafe fn modify(&mut self, range: FileRange) -> (FileRange, &mut [u8]) {
        let new = if let Some(&replaced) = self.replaced.get(&range.start) {
            assert_eq!(replaced.from, range);

            if let Some(to) = replaced.to {
                assert_eq!(to, range);
                to
            } else {
                let (to, _) = self.allocate(range.len());

                self.replaced.get_mut(&range.start).unwrap().to = Some(to);

                to
            }
        } else if let Some(&new) = self.new.get(&range.start) {
            assert_eq!(new, range);
            new
        } else {
            let (new, _) = self.allocate(range.len());

            let res = self.replaced.insert(
                new.start,
                Replaced {
                    from: range,
                    to: Some(new),
                },
            );
            debug_assert!(res.is_none());

            new
        };

        (new, self.get_data(new))
    }

    pub fn allocate(&mut self, length: u64) -> (FileRange, &mut [u8]) {
        unsafe {
            let new = self.allocate_range(length);

            let res = self.new.insert(new.start, new);
            debug_assert!(res.is_none());

            (new, self.get_data(new))
        }
    }

    pub fn free(&mut self, range: FileRange) {
        //
    }

    fn allocate_range(&mut self, length: u64) -> FileRange {
        let range: FileRange = todo!();

        range
    }
}

impl Db {
    fn header(&self) -> &Header {
        unsafe {
            self.reference_range_unchecked(Self::header_ptr().range(size_of::<Header>() as u64))
        }
    }

    fn header_mut(&mut self) -> &mut Header {
        unsafe { self.modify_range_unchecked(Self::header_ptr().range(size_of::<Header>() as u64)) }
    }

    pub fn add_slab(&mut self, size: u32) -> SlabPointer {
        let allocator_state = self.header().allocator_state;
        allocator_state.slabs.add_slab(self, size)
    }

    fn transaction(f: impl FnOnce(TransactionHandle)) {
        // let handle = TransactionHandle {};
    }

    fn header_ptr() -> FilePointer {
        FilePointer(0.into())
    }

    fn root_ptr() -> FilePointer {
        Self::header_ptr() + 16
    }

    fn allocator_state_ptr() -> FilePointer {
        Self::root_ptr() + size_of::<FilePointer>() as u64
    }

    fn general_purpose_allocator() -> GeneralPurposeAllocator {
        GeneralPurposeAllocator {
            head_ptr: Self::allocator_state_ptr(),
        }
    }

    pub fn create_reader(&self) -> Reader {
        Reader {
            state: self.state.clone(),
        }
    }

    fn update_root(&mut self, new_root: FilePointer) -> Arc<Snapshot> {
        // TODO: we could write some here + flush here for better consistency
        // e.g. a copy of the new root pointer

        // flush all data in file
        self.map.flush().unwrap();

        // update root pointer and immediately flush
        unsafe {
            self.write(Self::root_ptr(), new_root);
        }

        self.map
            .flush_range(Self::root_ptr().0.get() as usize, size_of::<FilePointer>())
            .unwrap();

        // update data that readers see
        self.state.swap(Arc::new(Snapshot {
            root: new_root,
            map: self.create_readonly_map(),
        }))
    }

    #[track_caller]
    unsafe fn read<T: FromBytes>(&self, at: FilePointer) -> T {
        self.read_range(at.range(size_of::<T>() as u64))
    }

    #[track_caller]
    unsafe fn read_range<T: FromBytes>(&self, range: FileRange) -> T {
        assert!(!range.start.is_null(), "null pointer dereference");
        LayoutVerified::<_, T>::new(&self.map[range.as_range()])
            .unwrap()
            .read()
    }

    #[track_caller]
    unsafe fn write<T: AsBytes>(&mut self, at: FilePointer, data: T) {
        self.write_range(at.range(size_of::<T>() as u64), data)
    }

    #[track_caller]
    unsafe fn write_range<T: AsBytes>(&mut self, range: FileRange, data: T) {
        assert!(!range.start.is_null(), "null pointer dereference");
        LayoutVerified::<_, T>::new(&mut self.map[range.as_range()])
            .unwrap()
            .write(data)
    }

    #[track_caller]
    unsafe fn modify<T: FromBytes + AsBytes>(&mut self, at: FilePointer) -> &mut T {
        self.modify_range(at.range(size_of::<T>() as u64))
    }

    #[track_caller]
    unsafe fn modify_range<T: FromBytes + AsBytes>(&mut self, range: FileRange) -> &mut T {
        assert!(!range.start.is_null(), "null pointer dereference");
        self.modify_range_unchecked(range)
    }

    unsafe fn modify_range_unchecked<T: FromBytes + AsBytes>(
        &mut self,
        range: FileRange,
    ) -> &mut T {
        LayoutVerified::<_, T>::new(&mut self.map[range.as_range()])
            .unwrap()
            .into_mut()
    }

    #[track_caller]
    unsafe fn reference<T: FromBytes>(&self, at: FilePointer) -> &T {
        self.reference_range(at.range(size_of::<T>() as u64))
    }

    #[track_caller]
    unsafe fn reference_range<T: FromBytes>(&self, range: FileRange) -> &T {
        assert!(!range.start.is_null(), "null pointer dereference");
        self.reference_range_unchecked(range)
    }

    unsafe fn reference_range_unchecked<T: FromBytes>(&self, range: FileRange) -> &T {
        LayoutVerified::<_, T>::new(&self.map[range.as_range()])
            .unwrap()
            .into_ref()
    }

    fn remap(&mut self) {
        let map = unsafe { MmapMut::map_mut(&self.file) }.unwrap();
        self.map = map;
    }

    fn create_readonly_map(&self) -> Mmap {
        unsafe { Mmap::map(&self.file) }.unwrap()
    }

    fn add_pages(&mut self, n: u64) -> PagePointer {
        // println!("adding {n} page{}", if n == 1 { "" } else { "s" });
        let len = self.file.metadata().unwrap().len();
        self.file.set_len(len + PAGE_SIZE * n).unwrap();
        self.remap();
        PagePointer::nth((len / PAGE_SIZE).try_into().unwrap())
    }

    pub fn new(file: File) -> Self {
        let len = file.metadata().unwrap().len();
        if len == 0 {
            file.set_len(PAGE_SIZE).unwrap();
        }

        let map = unsafe { MmapMut::map_mut(&file) }.unwrap();

        let mut db = Self {
            state: Arc::new(AtomicArc::new(Arc::new(Snapshot {
                root: FilePointer::null(),
                map: unsafe { Mmap::map(&file).unwrap() },
            }))),
            file,
            map,
        };

        unsafe {
            if len == 0 {
                *db.header_mut() = Header::default();
                db.init_allocator();
            }
        }

        let _ = db.state.swap(Arc::new(Snapshot {
            root: db.header().root,
            map: unsafe { Mmap::map(&db.file).unwrap() },
        }));

        db
    }

    unsafe fn init_allocator(&mut self) {
        let allocator_state = self.header().allocator_state;
        allocator_state.slabs.init(
            self,
            (PAGE_SIZE - size_of::<Header>() as u64).try_into().unwrap(),
        );
    }

    fn end_of_file(&self) -> FilePointer {
        FilePointer::null() + self.file.metadata().unwrap().len()
    }
}

#[cfg(test)]
mod tests {
    use std::io::Write;
    use std::process::Stdio;

    use rand::{thread_rng, Rng};

    use super::*;
    #[derive(Debug, Clone, Copy)]
    enum Operation {
        Allocate { size: u64 },
        Free { index: usize },
    }

    fn causes_fragmentation(sequence: &[Operation], print: bool) -> bool {
        let mut db = Db::new(tempfile::tempfile().unwrap());
        let allocator = Db::general_purpose_allocator();

        let mut ranges = Vec::new();

        for &operation in sequence {
            match operation {
                Operation::Allocate { size } => ranges.push(allocator.allocate(&mut db, size)),
                Operation::Free { index } => {
                    if ranges.get(index).is_some() {
                        allocator.free(&mut db, ranges.remove(index))
                    }
                }
            }
        }

        for range in ranges.drain(..) {
            allocator.free(&mut db, range);
        }

        fragmentation(&mut db, print) > 1
    }

    fn fragmentation(db: &mut Db, print: bool) -> usize {
        let allocator = Db::general_purpose_allocator();

        let mut next = unsafe { db.read::<FilePointer>(allocator.head_ptr) };

        let mut n = 0;
        while !next.is_null() {
            let size = GeneralPurposeAllocator::size(db, next);
            if print {
                println!("\x1b[34m[{n}]\x1b[m {:?}", next.range(size));
            }
            next = unsafe { db.read::<FilePointer>(next) };

            n += 1;
        }

        n
    }

    #[test]
    fn debug_fragmentation() {
        use Operation::*;
        #[rustfmt::skip]
        let mut sequence = vec![Allocate { size: 1946 }, Allocate { size: 3252 }, Free { index: 0 }, Allocate { size: 7391 }, Allocate { size: 394 }, Allocate { size: 3726 }, Allocate { size: 1429 }, Allocate { size: 3188 }, Allocate { size: 6375 }, Allocate { size: 4453 }, Allocate { size: 2514 }, Allocate { size: 4754 }, Allocate { size: 6785 }, Allocate { size: 2751 }, Allocate { size: 4107 }, Allocate { size: 3509 }, Allocate { size: 5897 }, Allocate { size: 7081 }, Allocate { size: 2419 }, Allocate { size: 5400 }, Allocate { size: 7135 }, Free { index: 14 }, Allocate { size: 2130 }, Free { index: 18 }, Allocate { size: 3450 }, Allocate { size: 1296 }, Allocate { size: 8091 }, Allocate { size: 4646 }, Allocate { size: 3891 }, Free { index: 0 }, Allocate { size: 1087 }, Allocate { size: 101 }, Allocate { size: 5353 }, Allocate { size: 3381 }, Allocate { size: 6869 }, Free { index: 1 }, Allocate { size: 3750 }, Allocate { size: 1398 }, Free { index: 22 }, Allocate { size: 18 }, Free { index: 25 }, Allocate { size: 642 }, Free { index: 4 }, Allocate { size: 4 }, Allocate { size: 1898 }, Allocate { size: 5259 }, Free { index: 26 }, Allocate { size: 3151 }, Allocate { size: 4989 }, Allocate { size: 6493 }, Allocate { size: 551 }, Allocate { size: 706 }, Allocate { size: 4161 }, Free { index: 16 }, Allocate { size: 3422 }, Allocate { size: 3011 }, Allocate { size: 5149 }, Allocate { size: 4687 }, Allocate { size: 5 }, Free { index: 34 }, Allocate { size: 191 }, Allocate { size: 2851 }, Allocate { size: 3597 }, Free { index: 28 }, Allocate { size: 7037 }, Allocate { size: 4660 }, Allocate { size: 194 }, Allocate { size: 5537 }, Allocate { size: 3242 }, Allocate { size: 6298 }, Allocate { size: 1239 }, Allocate { size: 7025 }, Allocate { size: 3563 }, Allocate { size: 5039 }, Free { index: 40 }, Allocate { size: 4549 }, Allocate { size: 5362 }, Allocate { size: 3510 }, Free { index: 31 }, Allocate { size: 226 }, Allocate { size: 6904 }, Allocate { size: 4150 }, Allocate { size: 4914 }, Allocate { size: 2330 }, Allocate { size: 2499 }, Allocate { size: 6677 }, Allocate { size: 95 }, Allocate { size: 3726 }, Allocate { size: 3258 }, Free { index: 2 }, Allocate { size: 2129 }, Allocate { size: 3674 }, Allocate { size: 1542 }, Allocate { size: 2210 }, Free { index: 21 }, Allocate { size: 3914 }, Allocate { size: 3108 }, Allocate { size: 1979 }, Allocate { size: 2677 }, Allocate { size: 8140 }, Allocate { size: 7573 }, Allocate { size: 121 }, Free { index: 59 }, Allocate { size: 6467 }, Allocate { size: 262 }, Allocate { size: 7711 }, Allocate { size: 2450 }, Allocate { size: 4351 }, Allocate { size: 4282 }, Free { index: 39 }, Allocate { size: 4050 }, Allocate { size: 67 }, Allocate { size: 5560 }, Free { index: 51 }, Allocate { size: 6038 }, Allocate { size: 555 }, Allocate { size: 1852 }, Free { index: 78 }, Allocate { size: 698 }];

        let mut prev_sequence = sequence.clone();
        while causes_fragmentation(&sequence, false) {
            prev_sequence = sequence.clone();
            sequence.pop();
        }

        println!("{prev_sequence:?}");

        let mut sequence = prev_sequence.clone();
        loop {
            let mut removed_something = false;
            let mut i = 0;
            while i < sequence.len() {
                let mut new_sequence = sequence.clone();

                new_sequence.remove(i);

                if causes_fragmentation(&new_sequence, false) {
                    removed_something = true;
                    println!("removed {i} ({:?})", sequence[i]);
                    sequence = new_sequence;
                } else {
                    for item in &mut new_sequence {
                        if let Operation::Free { index } = item {
                            if *index > i {
                                *index -= 1;
                            }
                        }
                    }

                    if causes_fragmentation(&new_sequence, false) {
                        removed_something = true;
                        println!("removed {i} ({:?}) after adjusting frees", sequence[i]);
                        sequence = new_sequence;
                    }
                }

                i += 1;
            }

            if !removed_something {
                break;
            }
        }

        loop {
            let mut merged_something = false;
            let mut i = 0;
            while i < sequence.len() {
                let mut new_sequence = sequence.clone();

                let removed = new_sequence.remove(i);

                if let Operation::Allocate { size: removed_size } = removed {
                    if let Some(Operation::Allocate { size }) = new_sequence.get_mut(i) {
                        *size += removed_size;
                    }
                }

                if causes_fragmentation(&new_sequence, false) {
                    merged_something = true;
                    println!(
                        "merged {} and {} ({:?} and {:?})",
                        i,
                        i + 1,
                        sequence[i],
                        sequence[i + 1]
                    );
                    sequence = new_sequence;
                } else {
                    for item in &mut new_sequence {
                        if let Operation::Free { index } = item {
                            if *index > i {
                                *index -= 1;
                            }
                        }
                    }

                    if causes_fragmentation(&new_sequence, false) {
                        merged_something = true;
                        println!(
                            "merged {} and {} ({:?} and {:?}) after adjusting frees",
                            i,
                            i + 1,
                            sequence[i],
                            sequence[i + 1]
                        );
                        sequence = new_sequence;
                    }
                }

                i += 1;
            }

            if !merged_something {
                break;
            }
        }

        println!("{sequence:?}");

        dbg!(causes_fragmentation(&sequence, true));
    }

    #[test]
    fn it_works() {
        let mut db = Db::new(tempfile::tempfile().unwrap());

        fn alloc_and_free_many<const N: usize>(db: &mut Db, n: u64) {
            let slab = db.add_slab(N as u32);

            let mut ranges = Vec::new();
            for i in 1..n {
                let range = slab.alloc(db);

                let data = [0; N].map(|_| i as u8);

                unsafe {
                    db.write_range(range, data);
                }

                ranges.push(range);
            }

            for range in ranges.into_iter().rev() {
                slab.free(db, range);
            }
        }

        // alloc_and_free_many::<1>(&mut db, 3 * PAGE_SIZE);
        alloc_and_free_many::<4>(&mut db, PAGE_SIZE / 4 * 3);
        alloc_and_free_many::<16>(&mut db, PAGE_SIZE / 16 * 3);

        let mut child = std::process::Command::new("hexdump")
            .arg("-C")
            .stdin(Stdio::piped())
            .stdout(Stdio::inherit())
            .spawn()
            .unwrap();

        let mut stdin = child.stdin.take().expect("failed to get stdin");

        stdin.write_all(db.map.as_bytes()).unwrap();

        std::mem::drop(stdin);

        child.wait().unwrap();
    }
}
